/*
 * C utility library
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include "library.h"
#include "smalloc.h"
#include "vector.h"
#include "cgi.h"
#include "rope.h"
#include "log.h"

#ifdef DEBUG
bool
vector_full(vector_t *v)
{
	if (v->length > v->size) {
		log("cgisys", PRI_ALERT, "Vector is over size: %d over %d",
						v->length, v->size);
		exit(0);
	}
	return (v->length == v->size);
}
#else
#define vector_full(v) (v->length == v->size)
#endif

vector_t *
vector_new(vector_delete_t func, int size)
{
	vector_t	*v;

	if (size <= 0) size = VECTOR_SIZE;

	v = (vector_t *)smalloc(sizeof(vector_t));
	v->cells = (vector_cell_t *)smalloc(size * sizeof(vector_cell_t));
	v->size = size;
	v->length = 0;
	v->c_delete = func;
	return v;
}

int
vector_length(vector_t *v)
{
	return v->length;
}

vector_cell_t
vector_data(vector_t *v, int offset)
{
	if (offset > v->length)
		return NULL;
	if (offset < 0)
		return NULL;
#if 0
	DPRINTF((stderr, "Vector data: Size %d, Length %d, Asked %d\n",
					v->size,
					v->length,
					offset));
#endif
	return v->cells[offset];
}

void
vector_delete(vector_t *v)
{
	int	 i;
	if (v->c_delete) {
		for (i = 0; i < vector_length(v); i++) {
			v->c_delete(vector_data(v, i));
		}
	}
	free(v->cells);
	free(v);
}

static inline vector_t *
vector_extend(vector_t *v)
{
	if (!vector_full(v)) return v;
	v->size *= 2;
	// DPRINTF((stderr, "Vector size is now %d\n", v->size));
	v->cells = srealloc(v->cells, v->size * sizeof(void *));
	return v;
}

static inline vector_t *
vector_shrink(vector_t *v)
{
	if (v->length < 9)
		return v;

	if ((v->length * 2) < v->size) {
		v->size /= 2;
		v->cells = (vector_cell_t *)
				srealloc(v->cells, v->size * sizeof(vector_cell_t));
	}
	return v;
}

void
vector_insert(vector_t *v, vector_cell_t val, int off)
{
	if (vector_full(v))
		vector_extend(v);

	if ((off < 0) || (off > v->length)) {
		DPRINTF((stderr, "Attempt to insert out of range (%d)\n", off));
		return;
	}
	
	if (off < v->length) {
		memmove(
				&(v->cells[(off + 1)]),
				&(v->cells[off]),
				((v->length - off) * sizeof(vector_cell_t))
				);
	}
	v->length++;	/* CAREFUL */
	if (v->length > 1000) {
		DPRINTF((stderr, "Vector length > 1000"));
	}
	v->cells[off] = val;
}

void
vector_push_back(vector_t *v, vector_cell_t val)
{
	vector_insert(v, val, v->length);
}

void
vector_push_front(vector_t *v, vector_cell_t val)
{
	vector_insert(v, val, 0);
}

vector_cell_t
vector_pop(vector_t *v, int off)
{
	vector_cell_t	 val;

	if ((off < 0) || (off > (v->length - 1))) {
		DPRINTF((stderr, "Attempt to pop out of range (%d)\n", off));
		return NULL;
	}

	val = v->cells[off];
	v->length--;
	if (off < v->length) {
		memmove(
				&(v->cells[off]),
				&(v->cells[(off + 1)]),
				((v->length - off) * sizeof(vector_cell_t))
				);
	}

	return val;
}

vector_cell_t
vector_pop_front(vector_t *v)
{
	return vector_pop(v, 0);
}

vector_cell_t
vector_pop_back(vector_t *v)
{
	return vector_pop(v, v->length - 1);
}

void
vector_remove(vector_t *v, int off)
{
	vector_cell_t	 val;
	val = vector_pop(v, off);
	if (v->c_delete)
		v->c_delete(val);
}

/* Note, compare_t is defined to compare void * */
int
vector_member(vector_t *v, compare_t func, void *val)
{
	int	 len;
	int	 i;

	len = vector_length(v);
	for (i = 0; i < len; i++) {
		if (func(val, vector_data(v, i)) == 0)
			return i;
	}
	return -1;
}
